"""Pi Hat EEPROM constants"""

from enum import IntEnum, IntFlag

__all__ = [
    'EepromSignature',
    'EepromVersion',
    'EepromAtomType',
    'EepromGpioDrive',
    'EepromGpioSlew',
    'EepromGpioHysteresis',
    'EepromGpioBackPower',
    'EepromGpioFunction',
    'EepromGpioPull',
]


class EepromSignature(IntFlag):
    """Magic signature"""

    RPI = 0x69502d52


class EepromVersion(IntFlag):
    """Structure version"""

    V1 = 1


class EepromAtomType(IntFlag):
    """Atom type"""

    INFO = 1
    GPIO = 2
    DTBO = 3
    CUSTOM = 4


class EepromGpioDrive(IntEnum):
    """GPIO bank drive"""

    DEFAULT = 0
    MA_2 = 1
    MA_4 = 2
    MA_6 = 3
    MA_8 = 4
    MA_10 = 5
    MA_12 = 6
    MA_14 = 7
    MA_16 = 8
    RESERVED_9 = 9
    RESERVED_10 = 10
    RESERVED_11 = 11
    RESERVED_12 = 12
    RESERVED_13 = 13
    RESERVED_14 = 14
    RESERVED_15 = 15


class EepromGpioSlew(IntEnum):
    """GPIO bank slew"""

    DEFAULT = 0
    LIMITED = 1
    UNLIMITED = 2
    RESERVED = 3


class EepromGpioHysteresis(IntEnum):
    """GPIO bank hysteresis"""

    DEFAULT = 0
    DISABLED = 1
    ENABLED = 2
    RESERVED = 3


class EepromGpioBackPower(IntEnum):
    """Board back powering"""

    NONE = 0
    MA_1300 = 1
    MA_2000 = 2
    RESERVED = 3


class EepromGpioFunction(IntEnum):
    """GPIO pin function"""

    INPUT = 0
    OUTPUT = 1
    ALT0 = 4
    ALT1 = 5
    ALT2 = 6
    ALT3 = 7
    ALT4 = 3
    ALT5 = 2


class EepromGpioPull(IntEnum):
    """GPIO pull direction"""

    DEFAULT = 0
    UP = 1
    DOWN = 2
    NONE = 3



"""CRC tests"""

from random import Random
import struct
import unittest
from pihat.eeprom.crc import crc16


class CrcTest(unittest.TestCase):
    """CRC tests"""

    def assertCrcEqual(self, data, crc):
        """Assert that CRC is equal to expected value"""
        self.assertEqual(crc16(data), crc)

    def assertCrcZero(self, data):
        """Assert that CRC is zero"""
        self.assertCrcEqual(data, 0)

    def assertCrcAppended(self, data):
        """Assert that appending CRC produces an overall CRC of zero"""
        self.assertCrcZero(data + struct.pack('<H', crc16(data)))

    def test_constants(self):
        """Test CRC of constant strings"""
        self.assertCrcEqual(b'', 0)
        self.assertCrcEqual(b'123456789', 0xbb3d)
        self.assertCrcEqual(b'Hello world!', 0x37b9)

    def test_append(self):
        """Test appending CRC to pseudo-random strings"""
        rng = Random(42)
        lengths = [rng.randint(0, 1024) for _ in range(10)]
        for length in lengths:
            data = bytes(rng.getrandbits(8) for _ in range(length))
            self.assertCrcAppended(data)



"""Pi Hat EEPROM exceptions"""

__all__ = [
    'EepromValueError',
    'EepromSignatureError',
    'EepromCrcError',
    'EepromLengthError',
]


class EepromValueError(ValueError):
    """EEPROM value error"""


class EepromSignatureError(EepromValueError):
    """EEPROM magic signature error"""


class EepromCrcError(EepromValueError):
    """EEPROM CRC error"""


class EepromLengthError(EepromValueError):
    """EEPROM length error"""



"""Pi Hat EEPROM layout"""

from abc import abstractmethod
from ctypes import (Array, LittleEndianStructure, c_uint8, c_uint16,
                    c_uint32, sizeof)
from dataclasses import dataclass, field
from typing import ClassVar, Dict, List, Optional, Type, Union
from uuid import UUID
from .constants import (EepromSignature, EepromVersion, EepromAtomType,
                        EepromGpioDrive, EepromGpioSlew, EepromGpioHysteresis,
                        EepromGpioBackPower, EepromGpioFunction,
                        EepromGpioPull)
from .crc import crc16
from .exceptions import EepromSignatureError, EepromLengthError, EepromCrcError

__all__ = [
    'EepromHeader',
    'EepromVendorInfo',
    'EepromGpioBank',
    'EepromGpioPower',
    'EepromGpioPin',
    'EepromGpioPins',
    'EepromGpioMap',
    'EepromAtom',
    'Eeprom',
]

CRC_SEED = 0
GPIO_COUNT = 28


class EepromStructure:
    """EEPROM data structure abstract base class"""

    def __bytes__(self):
        return self.pack()

    def __len__(self):
        return len(self.pack(fixup=False))

    def __eq__(self, other):
        if isinstance(self, type(other)):
            return self.pack(fixup=False) == other.pack(fixup=False)
        return NotImplemented

    def fixup(self):
        """Fix up fields for consistency"""

    def pack(self, fixup=True):
        """Pack structure to binary data"""
        if fixup:
            self.fixup()

    @classmethod
    @abstractmethod
    def unpack(cls, raw):
        """Unpack structure from binary data"""


@dataclass
class EepromTypedField:
    """A typed field within an EEPROM structure"""

    type: Type
    field: Optional[str] = None

    def __set_name__(self, owner, name):
        if self.field is None:
            self.field = '_%s' % name
        owner.register_typed_field(name)


@dataclass
class EepromEnumField(EepromTypedField):
    """An enumerated type field within an EEPROM structure"""

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self.type(getattr(instance, self.field))

    def __set__(self, instance, value):
        setattr(instance, self.field, self.type(value))


@dataclass
class EepromUuidField(EepromTypedField):
    """A UUID field within an EEPROM structure"""

    type: Type = UUID

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self.type(bytes=bytes(getattr(instance, self.field)))

    def __set__(self, instance, value):
        if isinstance(value, self.type):
            value = value.bytes
        ftype = type(getattr(instance, self.field))
        setattr(instance, self.field, ftype(*bytes(value)))


@dataclass
class EepromStringField(EepromTypedField):
    """A string field within an EEPROM structure"""

    type: Type = bytes
    length_field: Optional[str] = None

    def __set_name__(self, owner, name):
        super().__set_name__(owner, name)
        if self.length_field is None:
            self.length_field = '%s_len' % self.field

    def __get__(self, instance, owner):
        if instance is None:
            return self
        return self.type(getattr(instance, self.field))

    def __set__(self, instance, value):
        value = self.type(value)
        setattr(instance, self.field, value)
        setattr(instance, self.length_field, len(value))


class EepromLittleEndianStructure(EepromStructure, LittleEndianStructure):
    """An EEPROM data structure"""

    _typed_fields_: List[str] = []
    """List of attribute names for typed field attributes"""

    _repr_fields_: List[str] = []
    """List of attribute names to appear in `repr` output"""

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        cls._repr_fields_ = cls._typed_fields_ + [
            f[0] for f in getattr(cls, '_fields_', ())
            if not f[0].startswith('_')
        ]

    def __init__(self, **kwargs):
        typed_kwargs = {
            k: kwargs.pop(k) for k in self._typed_fields_ if k in kwargs
        }
        super().__init__(**kwargs)
        for k, v in typed_kwargs.items():
            setattr(self, k, v)

    def __repr__(self):
        return '%s(%s)' % (self.__class__.__name__, ', '.join(
            '%s=%r' % (f, getattr(self, f)) for f in self._repr_fields_
        ))

    def __len__(self):
        return sizeof(self)

    @classmethod
    def register_typed_field(cls, name):
        """Register a typed field attribute"""
        cls._typed_fields_ = cls._typed_fields_ + [name]

    def pack(self, fixup=True):
        """Pack structure to binary data"""
        super().pack(fixup=fixup)
        return bytes(memoryview(self))

    @classmethod
    def unpack(cls, raw):
        """Unpack structure from binary data"""
        try:
            return cls.from_buffer_copy(raw)
        except ValueError as exc:
            raise EepromLengthError(*exc.args)


class EepromArray(Array):
    """An EEPROM data structure array"""

    _type_: ClassVar[Type] = c_uint8
    _length_: ClassVar[int] = 0

    def __repr__(self):
        return '[%s]' % ', '.join('%r' % x for x in self)


class EepromHeader(EepromLittleEndianStructure):
    """EEPROM header"""

    _fields_ = [
        ('_signature', c_uint32),
        ('_version', c_uint8),
        ('_reserved', c_uint8),
        ('numatoms', c_uint16),
        ('eeplen', c_uint32),
    ]

    signature = EepromEnumField(EepromSignature)
    version = EepromEnumField(EepromVersion)


class EepromAtomData(EepromLittleEndianStructure):
    """EEPROM atom data"""

    type: ClassVar[EepromAtomType]

    types: ClassVar[Dict[EepromAtomType, Type]] = {}

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        if cls.type is not None:
            cls.types[cls.type] = cls


class EepromVendorInfo(EepromAtomData):
    """EEPROM vendor information data"""

    type: ClassVar[EepromAtomType] = EepromAtomType.INFO

    _vstr: bytes = b''
    _pstr: bytes = b''

    _fields_ = [
        ('_uuid', c_uint8 * 16),
        ('pid', c_uint16),
        ('pver', c_uint16),
        ('_vslen', c_uint8),
        ('_pslen', c_uint8),
    ]

    uuid = EepromUuidField()
    vstr = EepromStringField(length_field='_vslen')
    pstr = EepromStringField(length_field='_pslen')

    def __len__(self):
        return sizeof(self) + len(self._vstr) + len(self._pstr)

    def fixup(self):
        super().fixup()
        self._vslen = len(self.vstr)
        self._pslen = len(self.pstr)

    def pack(self, fixup=True):
        return super().pack(fixup=fixup) + self._vstr + self._pstr

    @classmethod
    def unpack(cls, raw):
        # pylint: disable=protected-access
        self = super().unpack(raw)
        hlen = sizeof(self)
        vslen = self._vslen
        pslen = self._pslen
        if len(raw) < hlen + vslen + pslen:
            raise EepromLengthError("Underlength vendor information data")
        self._vstr = raw[hlen:(hlen + vslen)]
        self._pstr = raw[(hlen + vslen):(hlen + vslen + pslen)]
        return self


class EepromGpioBank(EepromLittleEndianStructure):
    """EEPROM GPIO bank configuration"""

    _fields_ = [
        ('_drive', c_uint8, 4),
        ('_slew', c_uint8, 2),
        ('_hysteresis', c_uint8, 2),
    ]

    drive = EepromEnumField(EepromGpioDrive)
    slew = EepromEnumField(EepromGpioSlew)
    hysteresis = EepromEnumField(EepromGpioHysteresis)


class EepromGpioPower(EepromLittleEndianStructure):
    """EEPROM GPIO powering"""

    _fields_ = [
        ('_back_power', c_uint8, 2),
        ('_reserved', c_uint8, 6),
    ]

    back_power = EepromEnumField(EepromGpioBackPower)


class EepromGpioPin(EepromLittleEndianStructure):
    """EEPROM GPIO pin description"""

    _fields_ = [
        ('_function', c_uint8, 3),
        ('_reserved', c_uint8, 2),
        ('_pull', c_uint8, 2),
        ('_used', c_uint8, 1),
    ]

    function = EepromEnumField(EepromGpioFunction)
    pull = EepromEnumField(EepromGpioPull)
    used = EepromEnumField(bool)


class EepromGpioPins(EepromArray):
    """EEPROM GPIO pin descriptions"""

    _type_ = EepromGpioPin
    _length_ = GPIO_COUNT


class EepromGpioMap(EepromAtomData):
    """EEPROM GPIO map"""

    type: ClassVar[EepromAtomType] = EepromAtomType.GPIO

    _fields_ = [
        ('bank', EepromGpioBank),
        ('power', EepromGpioPower),
        ('pins', EepromGpioPins),
    ]


class EepromAtomChecksum(EepromLittleEndianStructure):
    """EEPROM atom checksum"""

    _fields_ = [
        ("crc", c_uint16),
    ]


class EepromAtom(EepromLittleEndianStructure):
    """EEPROM atom"""

    data: Union[bytes, EepromAtomData] = b''

    _typed_fields_ = ['data']

    _fields_ = [
        ('_type', c_uint16),
        ('count', c_uint16),
        ('_dlen', c_uint32),
    ]

    type = EepromEnumField(EepromAtomType)

    def __len__(self):
        return sizeof(self) + len(self.data) + sizeof(EepromAtomChecksum)

    def fixup(self):
        # pylint: disable=attribute-defined-outside-init
        super().fixup()
        self._dlen = len(self.data) + sizeof(EepromAtomChecksum)

    def pack(self, fixup=True):
        header = super().pack(fixup=fixup)
        data = (self.data.pack(fixup=fixup) if hasattr(self.data, 'pack')
                else bytes(self.data))
        body = header + data
        checksum = EepromAtomChecksum(crc=crc16(body, CRC_SEED))
        return body + checksum

    @classmethod
    def unpack(cls, raw):
        # pylint: disable=protected-access
        self = super().unpack(raw)
        hlen = sizeof(self)
        dlen = self._dlen
        clen = sizeof(EepromAtomChecksum)
        if dlen < clen:
            raise EepromLengthError("Underlength atom CRC")
        if len(raw) < hlen + dlen:
            raise EepromLengthError("Underlength atom data")
        atom = raw[:hlen + dlen]
        if crc16(atom, CRC_SEED) != 0:
            raise EepromCrcError(
                "Invalid atom CRC 0x%04x (expected 0x%04x)" % (
                    EepromAtomChecksum.unpack(atom[-clen:]).crc,
                    crc16(atom[:-clen])
                )
            )
        self.data = atom[hlen:-clen]
        subcls = EepromAtomData.types.get(self.type)
        if subcls is not None:
            self.data = subcls.unpack(self.data)
        return self


@dataclass
class EepromAtomAttribute:
    """An EEPROM attribute located inside an atom"""

    atom: str
    attribute: Optional[str] = None

    def __set_name__(self, owner, name):
        if self.attribute is None:
            self.attribute = name

    def __get__(self, instance, owner):
        if instance is None:
            return self
        atom = getattr(instance, self.atom)
        return getattr(atom.data, self.attribute)

    def __set__(self, instance, value):
        atom = getattr(instance, self.atom)
        setattr(atom.data, self.attribute, value)


@dataclass
class EepromInfoAttribute(EepromAtomAttribute):
    """An EEPROM attribute located inside the vendor information atom"""

    atom: str = 'info'


@dataclass
class EepromGpioAttribute(EepromAtomAttribute):
    """An EEPROM attribute located inside the GPIO map atom"""

    atom: str = 'gpio'


@dataclass
class Eeprom(EepromStructure):
    """EEPROM content"""

    header: EepromHeader = field(default_factory=lambda: EepromHeader(
        signature=EepromSignature.RPI, version=EepromVersion.V1,
    ))
    atoms: List[EepromAtom] = field(default_factory=lambda: [
        EepromAtom(type=EepromAtomType.INFO, count=0, data=EepromVendorInfo()),
        EepromAtom(type=EepromAtomType.GPIO, count=1, data=EepromGpioMap()),
    ])

    uuid = EepromInfoAttribute()
    pid = EepromInfoAttribute()
    pver = EepromInfoAttribute()
    vstr = EepromInfoAttribute()
    pstr = EepromInfoAttribute()

    bank = EepromGpioAttribute()
    power = EepromGpioAttribute()
    pins = EepromGpioAttribute()

    def __len__(self):
        return len(self.header) + sum(len(x) for x in self.atoms)

    def fixup(self):
        super().fixup()
        for i, atom in enumerate(self.atoms):
            atom.count = i
            atom.fixup()
        self.header.numatoms = len(self.atoms)
        self.header.eeplen = len(self)
        self.header.fixup()

    def pack(self, fixup=True):
        super().pack(fixup=fixup)
        atoms = [x.pack(fixup=False) for x in self.atoms]
        return self.header.pack(fixup=False) + b''.join(atoms)

    @classmethod
    def unpack(cls, raw):
        header = EepromHeader.unpack(raw)
        if header.signature != EepromSignature.RPI:
            raise EepromSignatureError("Invalid EEPROM signature")
        if len(raw) < header.eeplen:
            raise EepromLengthError("Underlength EEPROM content")
        remaining = raw[len(header):header.eeplen]
        atoms = []
        while remaining:
            atom = EepromAtom.unpack(remaining)
            atoms.append(atom)
            remaining = remaining[len(atom):]
        if header.numatoms != len(atoms):
            raise EepromLengthError("Atom count mismatch")
        return cls(header, atoms)

    def atom(self, type):
        """Find first atom of a specified type"""
        # pylint: disable=redefined-builtin
        return next((x for x in self.atoms if x.type == type), None)

    @property
    def info(self):
        """Vendor information atom"""
        atom = self.atom(EepromAtomType.INFO)
        if atom is None:
            atom = EepromAtom(type=EepromAtomType.INFO,
                              data=EepromVendorInfo())
            self.atoms.insert(0, atom)
        return atom

    @property
    def gpio(self):
        """GPIO map atom"""
        atom = self.atom(EepromAtomType.GPIO)
        if atom is None:
            atom = EepromAtom(type=EepromAtomType.GPIO, data=EepromGpioMap())
            self.atoms.insert(self.atoms.index(self.info) + 1, atom)
        return atom

    @property
    def dtbo(self):
        """Device tree overlay atom"""
        atom = self.atom(EepromAtomType.DTBO)
        if atom is None:
            atom = EepromAtom(type=EepromAtomType.DTBO)
            self.atoms.insert(self.atoms.index(self.gpio) + 1, atom)
        return atom



"""Layout tests"""

import struct
import unittest
from uuid import UUID
from pihat.eeprom import *


class HeaderTest(unittest.TestCase):
    """Header tests"""

    def test_signature(self):
        """Test signature field behaviour"""
        rpi = struct.unpack('<L', b'R-Pi')[0]
        header = EepromHeader()
        self.assertIsInstance(header.signature, EepromSignature)
        self.assertEqual(header.signature, 0)
        header.signature = EepromSignature.RPI
        self.assertIsInstance(header.signature, EepromSignature)
        self.assertEqual(header.signature, rpi)
        header.signature = rpi
        self.assertIsInstance(header.signature, EepromSignature)
        self.assertIs(header.signature, EepromSignature.RPI)
        header.signature = 0x12345678
        self.assertIsInstance(header.signature, EepromSignature)
        self.assertEqual(header.signature, 0x12345678)

    def test_version(self):
        """Test version field behaviour"""
        header = EepromHeader()
        self.assertIsInstance(header.version, EepromVersion)
        self.assertEqual(header.version, 0)
        header.version = EepromVersion.V1
        self.assertIsInstance(header.version, EepromVersion)
        self.assertEqual(header.version, 1)
        header.version = 1
        self.assertIsInstance(header.version, EepromVersion)
        self.assertIs(header.version, EepromVersion.V1)
        header.version = 42
        self.assertIsInstance(header.version, EepromVersion)
        self.assertEqual(header.version, 42)

    def test_pack(self):
        """Test packing to binary data"""
        header = EepromHeader(
            signature=EepromSignature.RPI, version=EepromVersion.V1,
            numatoms=2, eeplen=0x204,
        )
        raw = bytes([
            0x52, 0x2d, 0x50, 0x69, 0x01, 0x00, 0x02, 0x00,
            0x04, 0x02, 0x00, 0x00
        ])
        self.assertEqual(header.pack(), raw)
        self.assertEqual(bytes(header), raw)

    def test_unpack(self):
        """Test unpacking from binary data"""
        raw = bytes([
            0x52, 0x2d, 0x50, 0x69, 0x01, 0x00, 0x03, 0x00,
            0x27, 0x08, 0x00, 0x00
        ])
        header = EepromHeader.unpack(raw)
        self.assertIsInstance(header, EepromHeader)
        self.assertEqual(header.signature, EepromSignature.RPI)
        self.assertEqual(header.version, EepromVersion.V1)
        self.assertEqual(header.numatoms, 3)
        self.assertEqual(header.eeplen, 0x827)
        self.assertEqual(header, EepromHeader(
            signature=EepromSignature.RPI, version=EepromVersion.V1,
            numatoms=3, eeplen=0x827,
        ))
        with self.assertRaises(EepromLengthError):
            EepromHeader.unpack(raw[:-1])
        self.assertEqual(EepromHeader.unpack(raw + b'padding'), header)


class VendorInfoTest(unittest.TestCase):
    """Vendor information data tests"""

    def test_uuid(self):
        """Test UUID field behaviour"""
        info = EepromVendorInfo()
        self.assertIsInstance(info.uuid, UUID)
        self.assertEqual(info.uuid, UUID(int=0))
        info.uuid = UUID('aad2525e-01c2-40ed-b624-8b5a6c810987')
        self.assertIsInstance(info.uuid, UUID)
        self.assertEqual(info.uuid,
                         UUID('aad2525e-01c2-40ed-b624-8b5a6c810987'))
        info.uuid = bytes([0x6c, 0xe6, 0x44, 0x78, 0x8b, 0xf7, 0x43, 0x51,
                           0xb2, 0xd4, 0xe9, 0x67, 0xe1, 0x78, 0x0a, 0x5f])
        self.assertIsInstance(info.uuid, UUID)
        self.assertEqual(info.uuid,
                         UUID('6ce64478-8bf7-4351-b2d4-e967e1780a5f'))

    def test_str(self):
        """Test vendor and product string field behaviour"""
        info = EepromVendorInfo()
        self.assertEqual(len(info), 22)
        info.vstr = b'hello'
        self.assertEqual(len(info), 27)
        info.pstr = b'world'
        self.assertEqual(len(info), 32)
        info.vstr = b'hello '
        self.assertEqual(len(info), 33)
        self.assertEqual(info.vstr, b'hello ')
        self.assertEqual(info.pstr, b'world')
        self.assertEqual(bytes(info), b'\0' * 20 + b'\x06\x05hello world')

    def test_pack(self):
        """Test packing to binary data"""
        info = EepromVendorInfo(
            uuid=UUID('c475c256-54e6-4470-8418-268de3f48fb8'),
            pid=0x8086, pver=0xa0, vstr=b'Vendor', pstr=b'Product',
        )
        self.assertEqual(len(info), 35)
        raw = bytes([
            0xc4, 0x75, 0xc2, 0x56, 0x54, 0xe6, 0x44, 0x70,
            0x84, 0x18, 0x26, 0x8d, 0xe3, 0xf4, 0x8f, 0xb8,
            0x86, 0x80, 0xa0, 0x00, 0x06, 0x07
        ]) + b'VendorProduct'
        self.assertEqual(info.pack(), raw)
        self.assertEqual(bytes(info), raw)

    def test_unpack(self):
        """Test unpacking from binary data"""
        raw = bytes([
            0x9b, 0x60, 0xec, 0x55, 0x8f, 0xa3, 0x4c, 0xd4,
            0x9b, 0x01, 0xa6, 0xf1, 0x7b, 0x19, 0x03, 0xb3,
            0x34, 0x12, 0xcd, 0xab, 0x07, 0x05
        ]) + b'CompanyThing'
        info = EepromVendorInfo.unpack(raw)
        self.assertEqual(info.uuid,
                         UUID('9b60ec55-8fa3-4cd4-9b01-a6f17b1903b3'))
        self.assertEqual(info.pid, 0x1234)
        self.assertEqual(info.pver, 0xabcd)
        self.assertEqual(info.vstr, b'Company')
        self.assertEqual(info.pstr, b'Thing')
        with self.assertRaises(EepromLengthError):
            EepromVendorInfo.unpack(raw[:-1])
        self.assertEqual(EepromVendorInfo.unpack(raw + b'padding'), info)


class GpioMapTest(unittest.TestCase):
    """GPIO map tests"""

    def test_bank(self):
        """Test bank field behaviour"""
        gpio = EepromGpioMap()
        self.assertIsInstance(gpio.bank.drive, EepromGpioDrive)
        self.assertEqual(gpio.bank.drive, EepromGpioDrive.DEFAULT)
        self.assertIsInstance(gpio.bank.slew, EepromGpioSlew)
        self.assertEqual(gpio.bank.slew, EepromGpioSlew.DEFAULT)
        self.assertIsInstance(gpio.bank.hysteresis, EepromGpioHysteresis)
        self.assertEqual(gpio.bank.hysteresis, EepromGpioHysteresis.DEFAULT)
        self.assertEqual(bytes(gpio.bank), b'\0')
        gpio = EepromGpioMap(
            bank=EepromGpioBank(
                drive=EepromGpioDrive.MA_12,
                slew=EepromGpioSlew.LIMITED,
                hysteresis=EepromGpioHysteresis.ENABLED,
            ),
        )
        self.assertEqual(bytes(gpio.bank), bytes([0x96]))

    def test_power(self):
        """Test power field behaviour"""
        gpio = EepromGpioMap()
        self.assertIsInstance(gpio.power.back_power, EepromGpioBackPower)
        self.assertEqual(gpio.power.back_power, EepromGpioBackPower.NONE)
        self.assertEqual(bytes(gpio.power), b'\0')
        gpio = EepromGpioMap(
            power=EepromGpioPower(back_power=EepromGpioBackPower.MA_1300)
        )
        self.assertEqual(bytes(gpio.power), bytes([0x01]))

    def test_pins(self):
        """Test pins field behaviour"""
        gpio = EepromGpioMap()
        pin = gpio.pins[7]
        self.assertIsInstance(pin.used, bool)
        self.assertFalse(pin.used)
        self.assertIsInstance(pin.function, EepromGpioFunction)
        self.assertEqual(pin.function, EepromGpioFunction.INPUT)
        self.assertIsInstance(pin.pull, EepromGpioPull)
        self.assertEqual(pin.pull, EepromGpioPull.DEFAULT)
        gpio = EepromGpioMap(
            pins=EepromGpioPins(*(
                [EepromGpioPin()] * 11 +
                [EepromGpioPin(used=True, function=EepromGpioFunction.ALT5)]
            ))
        )
        self.assertEqual(
            bytes(gpio.pins),
            b'\0\0\0\0\0\0\0\0\0\0\0\x82\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0\0'
        )

    def test_pack(self):
        """Test packing to binary data"""
        gpio = EepromGpioMap(
            bank=EepromGpioBank(
                drive=EepromGpioDrive.MA_4,
                hysteresis=EepromGpioHysteresis.ENABLED,
            ),
            power=EepromGpioPower(
                back_power=EepromGpioBackPower.MA_2000,
            ),
            pins=EepromGpioPins(
                EepromGpioPin(),
                EepromGpioPin(),
                EepromGpioPin(
                    used=True,
                    function=EepromGpioFunction.ALT2,
                    pull=EepromGpioPull.DOWN,
                ),
                EepromGpioPin(
                    used=True,
                    function=EepromGpioFunction.ALT4,
                ),
            ),
        )
        self.assertEqual(len(gpio), 30)
        raw = bytes([
            0x82, 0x02, 0, 0, 0xc6, 0x83, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
        ])
        self.assertEqual(gpio.pack(), raw)
        self.assertEqual(bytes(gpio), raw)

    def test_unpack(self):
        """Test unpacking from binary data"""
        raw = bytes([
            0x03, 0x01, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0x81,
            0, 0, 0, 0, 0xc6, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0xa7
        ])
        gpio = EepromGpioMap.unpack(raw)
        self.assertEqual(gpio.bank.drive, EepromGpioDrive.MA_6)
        self.assertEqual(gpio.power.back_power, EepromGpioBackPower.MA_1300)
        self.assertFalse(gpio.pins[11].used)
        self.assertTrue(gpio.pins[12].used)
        self.assertEqual(gpio.pins[12].function, EepromGpioFunction.OUTPUT)
        self.assertTrue(gpio.pins[27].used)
        self.assertEqual(gpio.pins[27].pull, EepromGpioPull.UP)
        self.assertEqual(gpio.pins[27].function, EepromGpioFunction.ALT3)
        with self.assertRaises(EepromLengthError):
            EepromGpioMap.unpack(raw[:-1])
        self.assertEqual(EepromGpioMap.unpack(raw + b'padding'), gpio)


class AtomTest(unittest.TestCase):
    """Atom tests"""

    def test_pack(self):
        """Test packing to binary data"""
        atom = EepromAtom(
            type=EepromAtomType.INFO,
            count=2,
            data=EepromVendorInfo(
                uuid=UUID('fa421d59-c8dc-483b-b773-d2112194fc60'),
                pid=0xf00d,
                vstr=b'Vendor',
            )
        )
        self.assertEqual(len(atom), 38)
        raw = bytes([
            0x01, 0x00, 0x02, 0x00, 0x1e, 0x00, 0x00, 0x00,
            0xfa, 0x42, 0x1d, 0x59, 0xc8, 0xdc, 0x48, 0x3b,
            0xb7, 0x73, 0xd2, 0x11, 0x21, 0x94, 0xfc, 0x60,
            0x0d, 0xf0, 0x00, 0x00, 0x06, 0x00, 0x56, 0x65,
            0x6e, 0x64, 0x6f, 0x72, 0x3c, 0x99
        ])
        self.assertEqual(crc16(raw), 0)
        self.assertEqual(atom.pack(), raw)
        self.assertEqual(bytes(atom), raw)

    def test_unpack(self):
        """Test unpacking from binary data"""
        raw = bytes([
            0x01, 0x00, 0x03, 0x00, 0x1a, 0x00, 0x00, 0x00,
            0x4c, 0x4b, 0xa1, 0x2f, 0xb7, 0xe5, 0x44, 0x74,
            0x8a, 0xd4, 0xd4, 0xa0, 0xe1, 0xea, 0xa2, 0x6e,
            0xce, 0xfa, 0xfe, 0xca, 0x00, 0x02, 0x49, 0x74,
            0x7a, 0x54
        ])
        atom = EepromAtom.unpack(raw)
        self.assertEqual(atom.type, EepromAtomType.INFO)
        self.assertEqual(atom.count, 3)
        self.assertIsInstance(atom.data, EepromVendorInfo)
        self.assertEqual(atom.data.uuid,
                         UUID('4c4ba12f-b7e5-4474-8ad4-d4a0e1eaa26e'))
        self.assertEqual(atom.data.pid, 0xface)
        self.assertEqual(atom.data.pver, 0xcafe)
        self.assertEqual(atom.data.vstr, b'')
        self.assertEqual(atom.data.pstr, b'It')
        with self.assertRaises(EepromLengthError):
            EepromAtom.unpack(raw[:-1])
        self.assertEqual(EepromAtom.unpack(raw + b'padding'), atom)
        with self.assertRaises(EepromCrcError):
            EepromAtom.unpack(raw[:-1] + b'\0')
        raw = bytes([
            0x2a, 0x00, 0x01, 0x00, 0x07, 0x00, 0x00, 0x00,
            0x48, 0x65, 0x6c, 0x6c, 0x6f, 0x16, 0x55
        ])
        atom = EepromAtom.unpack(raw)
        self.assertIsInstance(atom.type, EepromAtomType)
        self.assertEqual(atom.type, 42)
        self.assertEqual(atom.count, 1)
        self.assertEqual(atom.data, b'Hello')


class EepromTest(unittest.TestCase):
    """EEPROM content tests"""

    def test_proxy(self):
        """Test attribute proxying"""
        eeprom = Eeprom()
        eeprom.uuid = UUID('d97855ff-d424-43ad-9a32-4fa04d1867ad')
        eeprom.pid = 0x8086
        eeprom.pver = 0xb3
        eeprom.vstr = b'Maker'
        eeprom.pstr = b'Thing'
        eeprom.bank.drive = EepromGpioDrive.MA_8
        eeprom.power.back_power = EepromGpioBackPower.MA_2000
        eeprom.pins[4].used = True
        eeprom.pins[4].function = EepromGpioFunction.ALT3
        info = eeprom.info
        gpio = eeprom.gpio
        self.assertEqual(eeprom.atoms, [info, gpio])
        self.assertEqual(info.type, EepromAtomType.INFO)
        self.assertIsInstance(info.data, EepromVendorInfo)
        self.assertEqual(gpio.type, EepromAtomType.GPIO)
        self.assertIsInstance(gpio.data, EepromGpioMap)
        self.assertEqual(bytes(info.data), bytes([
            0xd9, 0x78, 0x55, 0xff, 0xd4, 0x24, 0x43, 0xad,
            0x9a, 0x32, 0x4f, 0xa0, 0x4d, 0x18, 0x67, 0xad,
            0x86, 0x80, 0xb3, 0x00, 0x05, 0x05
        ]) + b'MakerThing')
        self.assertEqual(bytes(gpio.data), bytes([
            0x04, 0x02, 0, 0, 0, 0, 0x87, 0, 0, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
        ]))

    def test_autocreate(self):
        """Test automatic creation of atoms"""
        eeprom = Eeprom(atoms=[])
        self.assertEqual(eeprom.atoms, [])
        info = eeprom.info
        self.assertEqual(eeprom.atoms, [info])
        self.assertEqual(info.type, EepromAtomType.INFO)
        self.assertIsInstance(info.data, EepromVendorInfo)
        gpio = eeprom.gpio
        self.assertEqual(eeprom.atoms, [info, gpio])
        self.assertEqual(gpio.type, EepromAtomType.GPIO)
        self.assertIsInstance(gpio.data, EepromGpioMap)
        dtbo = eeprom.dtbo
        self.assertEqual(eeprom.atoms, [info, gpio, dtbo])
        self.assertEqual(dtbo.type, EepromAtomType.DTBO)
        eeprom = Eeprom(atoms=[])
        gpio = eeprom.gpio
        dtbo = eeprom.dtbo
        info = eeprom.info
        self.assertEqual(eeprom.atoms, [info, gpio, dtbo])
        eeprom.fixup()
        self.assertEqual(info.count, 0)
        self.assertEqual(gpio.count, 1)
        self.assertEqual(dtbo.count, 2)
        self.assertEqual(eeprom.header.numatoms, 3)

    def test_pack_unpack(self):
        """Test packing and unpacking of binary data"""
        eeprom = Eeprom()
        eeprom.uuid = UUID('717fa646-cf30-4529-9c5e-a17bb8cec598')
        eeprom.pstr = b'Object'
        eeprom.pins[7].used = True
        eeprom.pins[7].pull = EepromGpioPull.UP
        self.assertEqual(len(eeprom), 90)
        raw = bytes([
            0x52, 0x2d, 0x50, 0x69, 0x01, 0x00, 0x02, 0x00,
            0x5a, 0x00, 0x00, 0x00,
            0x01, 0x00, 0x00, 0x00, 0x1e, 0x00, 0x00, 0x00,
            0x71, 0x7f, 0xa6, 0x46, 0xcf, 0x30, 0x45, 0x29,
            0x9c, 0x5e, 0xa1, 0x7b, 0xb8, 0xce, 0xc5, 0x98,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x06, 0x4f, 0x62,
            0x6a, 0x65, 0x63, 0x74, 0x1a, 0x53,
            0x02, 0x00, 0x01, 0x00, 0x20, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0xa0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
            0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x55, 0xc4,
        ])
        self.assertEqual(eeprom.pack(), raw)
        self.assertEqual(bytes(eeprom), raw)
        self.assertEqual(Eeprom.unpack(raw), eeprom)
        with self.assertRaises(EepromSignatureError):
            Eeprom.unpack(b'x' + raw[1:])
        with self.assertRaises(EepromLengthError):
            Eeprom.unpack(raw[:-1])
        self.assertEqual(Eeprom.unpack(raw + b'padding'), eeprom)
        with self.assertRaises(EepromCrcError):
            Eeprom.unpack(raw[:-1] + b'\0')



"""CRC-16 algorithm"""

import array
from itertools import repeat

__all__ = ['crc16']

CRC16 = 0xa001


def crctab(poly):
    """Generate CRC lookup table"""
    table = array.array('H', repeat(0, 0x100))
    crc = 0x0001
    i = 0x80
    while i:
        crc = (crc >> 1) ^ (poly if crc & 1 else 0)
        for j in range(0, 0x100, 2 * i):
            table[i ^ j] = crc ^ table[j]
        i >>= 1
    return table


def crc16(data, crc=0, table=crctab(CRC16)):
    """Calculate CRC-16"""
    for x in data:
        crc = (crc >> 8) ^ table[(x ^ crc) & 0xff]
    return crc



#!/usr/bin/env python3

"""Setup script"""

from setuptools import setup, find_packages

setup(
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: "
        "GNU General Public License v2 or later (GPLv2+)",
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Embedded Systems",
        "Topic :: Software Development :: Libraries",
        "Topic :: System :: Boot",
        "Topic :: System :: Hardware :: Hardware Drivers",
    ],
    packages=find_packages(),
)
